# -*- coding: utf-8 -*-
import shutil
import tempfile
import unittest
from pathlib import Path

from htrmopo import get_listing, get_description, get_model
from htrmopo.record import v0RepositoryRecord, v1RepositoryRecord

thisfile = Path(__file__).resolve().parent
resources = thisfile / 'resources'


def _check_file_in_model_dir(dir_path, file_name, size):
    for file in dir_path.iterdir():
        if file.name == file_name and file.stat().st_size == size:
            return True
    return False


class TestQuery(unittest.TestCase):
    """
    Testing querying the model repository.
    """

    def setUp(self):
        self.temp_model = tempfile.TemporaryDirectory()
        self.temp_path = Path(self.temp_model.name)

    def tearDown(self):
        shutil.rmtree(self.temp_model.name)

    def test_listing(self):
        """
        Tests fetching the model list.
        """
        records = get_listing()
        self.assertGreater(len(records), 15)

    def test_get_description_v0(self):
        """
        Tests fetching the description of a v0 model.
        """
        record = get_description('10.5281/zenodo.8425684')
        self.assertIsInstance(record, v0RepositoryRecord)
        self.assertEqual(record.doi, '10.5281/zenodo.8425684')

    def test_get_description_concept_doi_v0(self):
        """
        Tests fetching the description of a v0 model by concept DOI.
        """
        record = get_description('10.5281/zenodo.8425683')
        self.assertIsInstance(record, v0RepositoryRecord)
        self.assertEqual(record.concept_doi, '10.5281/zenodo.8425683')
        self.assertNotEqual(record.doi, '10.5281/zenodo.8425683')

    def test_get_description_prev_version_v0(self):
        """
        Tests fetching the description of a v0 model that has a superseding newer version.
        """
        record = get_description('10.5281/zenodo.6657809')
        self.assertIsInstance(record, v0RepositoryRecord)
        self.assertEqual(record.doi, '10.5281/zenodo.6657809')
        self.assertEqual(record.concept_doi, '10.5281/zenodo.6657808')

    def test_get_description_v1(self):
        """
        Tests fetching the description of a v1 model.
        """
        record = get_description('10.5281/zenodo.14585602')
        self.assertIsInstance(record, v1RepositoryRecord)
        self.assertEqual(record.doi, '10.5281/zenodo.14585602')

    def test_get_description_concept_doi_v1(self):
        """
        Tests fetching the description of a v1 model by concept DOI.
        """
        record = get_description('10.5281/zenodo.7051645')
        self.assertIsInstance(record, v1RepositoryRecord)
        self.assertEqual(record.concept_doi, '10.5281/zenodo.7051645')
        self.assertNotEqual(record.doi, '10.5281/zenodo.7051645')

    def test_get_model_v0(self):
        """
        Tests fetching a v0 model.
        """
        path = get_model('10.5281/zenodo.8425684', path=self.temp_model.name)
        _check_file_in_model_dir(path, 'urdu_best.mlmodel', 16280146)

    def test_get_model_concept_doi_v0(self):
        """
        Tests fetching a v0 model by concept DOI.
        """
        path = get_model('10.5281/zenodo.8425683', path=self.temp_model.name)
        _check_file_in_model_dir(path, 'urdu_best.mlmodel', 16280146)

    def test_get_model_v1(self):
        """
        Tests fetching a v1 model.
        """
        path = get_model('10.5281/zenodo.14585602', path=self.temp_model.name)
        _check_file_in_model_dir(path, 'urdu_best.mlmodel', 16280146)

    def test_get_model_concept_doi_v1 (self):
        """
        Tests fetching a v1 model by concept DOI.
        """
        path = get_model('10.5281/zenodo.7051645', path=self.temp_model.name)
        _check_file_in_model_dir(path, 'urdu_best.mlmodel', 16280146)

    def test_prev_record_version_get_model_v0(self):
        """
        Tests fetching a model that has a superseding newer version.
        """
        path = get_model('10.5281/zenodo.6657809', path=self.temp_model.name)
        _check_file_in_model_dir(path, 'HTR-United-Manu_McFrench.mlmodel', 16176844)



#
# Copyright 2024 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
Accessors to the v2 model repository on zenodo.
"""
import yaml
import uuid
import logging
import requests

from collections import defaultdict
from contextlib import closing
from pathlib import Path
from sickle import Sickle
from urllib.parse import urlsplit
from platformdirs import user_data_dir
from jsonschema import validate
from dateutil.parser import parse as date_parse

from typing import TYPE_CHECKING, Any, Callable, Optional, Literal, Union, Dict


from htrmopo.record import DCATRecord, v0RepositoryRecord, v1RepositoryRecord
from htrmopo.util import (_yaml_regex, _v1_schema, _v0_schema, _doi_to_oai_id,
                          _doi_to_zenodo_id, get_repo_url, get_oai_url,
                          format_checker)

if TYPE_CHECKING:
    from os import PathLike

__all__ = ['get_model', 'get_description', 'get_listing']

logger = logging.getLogger(__name__)

sickle = Sickle(get_oai_url())
sickle.class_mapping['ListRecords'] = DCATRecord
sickle.class_mapping['GetRecord'] = DCATRecord


def _build_v0_record(metadata, request):
    try:
        json_metadata = request.json()
    except Exception:
        raise Exception(f'Metadata for {metadata["doi"]} not in JSON format')
    validate(json_metadata, _v0_schema, format_checker=format_checker)
    metadata.update(json_metadata)
    metadata['creators'].extend(metadata['authors'])
    metadata['doi'] = urlsplit(metadata['doi']).path[1:]
    metadata['concept_doi'] = urlsplit(metadata['concept_doi']).path[1:]
    metadata['metrics'] = {'cer': (100 - float(metadata.pop('accuracy')))}
    metadata.pop('authors')
    metadata.pop('name')
    return v0RepositoryRecord(**metadata)


def _build_v1_record(metadata, model_card):
    header, content = _yaml_regex.match(model_card).groups()
    try:
        yaml_metadata = yaml.safe_load(header)
    except Exception:
        raise Exception(f'Metadata for {metadata["doi"]} not in YAML format')
    validate(yaml_metadata, _v1_schema, format_checker=format_checker)
    metadata.update(yaml_metadata)
    if metadata['license'].startswith('other'):
        metadata['license'] = metadata['license_name']
        metadata.pop('license_name')
    # merge authors and creators
    metadata['creators'].extend(metadata['authors'])
    metadata['doi'] = urlsplit(metadata['doi']).path[1:]
    metadata['concept_doi'] = urlsplit(metadata['concept_doi']).path[1:]
    metadata.pop('id')
    metadata.pop('authors')
    metadata.pop('description', None)
    metadata['keywords'] = list({*metadata.pop('keywords', None), *metadata.pop('tags', None)})
    return v1RepositoryRecord(description=content, **metadata)


def get_model(model_id: str,
              path: Optional['PathLike'] = None,
              callback: Callable[[int, int], Any] = lambda total, advance: None,
              abort_if_exists: bool = False) -> 'PathLike':
    """
    Retrieves a model and their files to a path.

    Args:
        model_id: DOI of the model
        path: Optional destination path to write model to. If none is given a
              new one will be created in the htrmopo data dir.
        callback: Function called for every 1024 octet chunk received.
        version: Which version of the metadata should be returned if more than
                 one exists. Per default the newest is returned.

    Returns:
        The output path under which the model files have been placed.
    """
    logger.info(f'Saving model {model_id} to {path}')
    if (oai_id := _doi_to_oai_id(model_id)) is None:
        raise ValueError(f'{model_id} is not a valid DOI')
    try:
        record = sickle.GetRecord(identifier=oai_id, metadataPrefix='dcat')
    except requests.HTTPError:
        # might just be a concept DOI which don't show up in OAI. Try to
        # resolve it through the search API.
        r = requests.get(f'{get_repo_url()}records/{_doi_to_zenodo_id(model_id)}')
        r.raise_for_status()
        model_id = r.json()['doi']
        real_oai_id = _doi_to_oai_id(model_id)
        record = sickle.GetRecord(identifier=real_oai_id, metadataPrefix='dcat')

    if path is not None:
        path = Path(path).resolve()
    else:
        path = Path(user_data_dir('htrmopo')) / str(uuid.uuid5(uuid.NAMESPACE_DNS, model_id))

    if path.exists():
        if abort_if_exists:
            raise ValueError(f'Output path {path} already exists.')
        else:
            logger.warning(f'Output path {path} already exists.')

    path.mkdir(parents=True, exist_ok=True)

    callback(0, 0)
    metadata = record.metadata
    model_size = sum([file['size'] for file in metadata['distribution']])
    for file in metadata['distribution']:
        name = Path(urlsplit(file['url']).path).name
        output_path = (path / name).resolve()
        if not output_path.is_relative_to(path):
            logger.warning(f'Found file {name} with invalid path.')
            continue
        logger.info(f'Downloading model file {file["url"]} to {output_path}')
        with closing(requests.get(file["url"], stream=True)) as r:
            with open(output_path, 'wb') as f:
                for chunk in r.iter_content(chunk_size=1024):
                    callback(model_size, len(chunk))
                    f.write(chunk)
    return path


def get_description(model_id: str,
                    callback: Callable[..., Any] = lambda: None,
                    version: Optional[Literal['v0', 'v1']] = None) -> Union[v0RepositoryRecord, v1RepositoryRecord]:
    """
    Fetches the metadata for a single model from the zenodo repository.

    Args:
        model_id: DOI of the model.
        callback: Optional function called once per HTTP request.
        version: Which version of the metadata should be returned if more than
                 one exists. Per default the newest is returned.

    Returns:
        Dict
    """
    logger.info(f'Retrieving metadata for {model_id}')
    if (oai_id := _doi_to_oai_id(model_id)) is None:
        raise ValueError(f'{model_id} is not a valid DOI')
    try:
        record = sickle.GetRecord(identifier=oai_id, metadataPrefix='dcat')
    except requests.HTTPError:
        # might just be a concept DOI which don't show up in OAI. Try to
        # resolve it through the search API.
        r = requests.get(f'{get_repo_url()}records/{_doi_to_zenodo_id(model_id)}')
        r.raise_for_status()
        real_oai_id = _doi_to_oai_id(r.json()['doi'])
        record = sickle.GetRecord(identifier=real_oai_id, metadataPrefix='dcat')

    callback()
    repo_record = None
    for file in record.metadata['distribution']:
        # Zenodo deposits do not contain directories so we can just check for
        # file name.
        name = Path(urlsplit(file['url']).path).name
        if name == 'README.md':  # v1 model
            if version == 'v0':
                continue
            callback()
            r = requests.get(file['url'])
            r.raise_for_status()
            metadata = record.metadata.copy()
            metadata['publication_date'] = date_parse(record.header.datestamp)
            try:
                repo_record = _build_v1_record(metadata, r.content.decode('utf-8'))
            except Exception as e:
                logger.info(f'Invalid metadata for {model_id}: {e}')
        if name == 'metadata.json':  # v0 model
            if version != 'v0' and repo_record:
                break
            callback()
            r = requests.get(file['url'])
            r.raise_for_status()
            metadata = record.metadata.copy()
            metadata['publication_date'] = date_parse(record.header.datestamp)
            try:
                repo_record = _build_v0_record(metadata, r)
            except Exception as e:
                logger.info(f'Invalid metadata for {model_id}: {e}')

    if not repo_record:
        raise ValueError(f"No metadata found for \'{model_id}\'")
    return repo_record


def get_listing(callback: Callable[[int, int], Any] = lambda total, advance: None,
                from_date: Optional[str] = None) -> Dict[str, Dict[str, Union[v0RepositoryRecord, v1RepositoryRecord]]]:
    """
    Fetches a listing of all models from the zenodo repository.

    Args:
        callback: Function called for each processed record.
        from_date: Lower limit of time stamp to harvest.

    Returns:
        Dict of models with each model.
    """
    items = defaultdict(dict)
    # get all identiifers first just to get total number of items in repository
    total = len(list(sickle.ListIdentifiers(metadataPrefix='dcat', set='user-ocr_models', **{'from': from_date})))
    callback(total, 0)
    for record in sickle.ListRecords(metadataPrefix='dcat', set='user-ocr_models', **{'from': from_date}):
        doi = urlsplit(record.metadata['doi']).path[1:]
        for file_url in record.metadata['distribution']:
            if file_url['url'].endswith('README.md'):  # v1 model
                r = requests.get(file_url['url'])
                r.raise_for_status()
                metadata = record.metadata.copy()
                metadata['publication_date'] = date_parse(record.header.datestamp)
                try:
                    items[doi]['v1'] = _build_v1_record(metadata, r.content.decode('utf-8'))
                except Exception as e:
                    logger.info(f'Invalid metadata for {doi}: {e}')
            if file_url['url'].endswith('metadata.json'):  # v0 model
                r = requests.get(file_url['url'])
                r.raise_for_status()
                metadata = record.metadata.copy()
                metadata['publication_date'] = date_parse(record.header.datestamp)
                try:
                    items[doi]['v0'] = _build_v0_record(metadata, r)
                except Exception as e:
                    logger.info(f'Invalid metadata for {doi}: {e}')
        callback(total, 1)
    return dict(items)



#
# Copyright 2024 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
Accessors to the v2 model repository on zenodo.
"""
import yaml
import json
import logging
import requests

from pathlib import Path
from markdown import markdown
from jsonschema import validate
from typing import TYPE_CHECKING, Any, Callable

from htrmopo.util import (_yaml_regex, _v1_schema, _doi_to_zenodo_id,
                          get_repo_url, format_checker)

if TYPE_CHECKING:
    from os import PathLike

__all__ = ['update_model', 'publish_model']

logger = logging.getLogger(__name__)


def update_model(model_id: str,
                 model: 'PathLike',
                 model_card: str,
                 access_token: str,
                 callback: Callable[[int, int], Any] = lambda total, advance: None,
                 private: bool = False) -> str:
    """
    Updates a model in the repository by creating a new version of an existing
    record.

    Args:
        model_id: Model (NOT concept ID) of the record to create a new version
                  for.
        model: Path to read model from. Can be a single file or a directory.
        model_card: String of a Markdown document with YAML front matter
                    containing the model card + associated metadata.
        access_token: Zenodo API access token
        callback: Function called with octet-wise progress.
        private: Whether to generate a community inclusion request that makes
                 the model recoverable by the public.
    """
    model = Path(model)
    if not model.exists():
        raise ValueError('Model path {model} does not exist.')
    # find all files for the deposition
    model_size = 4
    if model.is_dir():
        model_files = []
        for file in model.iterdir():
            if file.relative_to(model) == Path('README.md'):
                logger.warning('Found a README.md in root of model dir. Ignoring.')
                continue
            if file.is_dir():
                logger.warning(f'Zenodo depositions are flat. Found dir {file}. Skipping')
                continue
            model_files.append(file)
            model_size += file.stat().st_size
    else:
        model_files = [model]
        model_size += model.stat().st_size

    logger.info(f'Got {len(model_files)} files for model with size of {model_size} bytes.')
    # validate model card
    header, content = _yaml_regex.match(model_card).groups()
    mopo_metadata = yaml.safe_load(header)

    # we first fetch the metadata record as the DOI argument might refer to a
    # concept ID which needs to be resolved.
    recid = _doi_to_zenodo_id(model_id)
    r = requests.get(f'{get_repo_url()}records/{recid}')
    r.raise_for_status()
    new_recid = _doi_to_zenodo_id(r.json()['doi'])
    if recid != new_recid:
        logger.info(f'Resolved {recid} to {new_recid}.')
        recid = new_recid
    callback(model_size, 1)

    # create a new draft version of an existing deposition.
    r = requests.post(f'{get_repo_url()}records/{recid}/versions', params={'access_token': access_token})
    r.raise_for_status()
    callback(model_size, 1)
    new_recid = r.json()['id']
    r = requests.get(f'{get_repo_url()}deposit/depositions/{new_recid}', params={'access_token': access_token})
    r.raise_for_status()
    callback(model_size, 1)
    depo = r.json()
    bucket_url = depo['links']['bucket']

    # finalize metadata record
    mopo_metadata['id'] = depo['metadata']['prereserve_doi']['doi']
    # we need to validate here to as the data structure might have an empty ID before
    validate(mopo_metadata, _v1_schema, format_checker=format_checker)

    header = yaml.dump(mopo_metadata)

    _metadata = f'---\n{header}---\n{content}'.encode('utf-8')
    model_size += len(_metadata)

    # push metadata file
    r = requests.put(f'{bucket_url}/README.md',
                     params={'access_token': access_token},
                     data=_metadata)
    r.raise_for_status()
    callback(model_size, len(_metadata))
    # push all other files
    for file in model_files:
        with open(file, 'rb') as fp:
            r = requests.put(f'{bucket_url}/{file.name}',
                             params={'access_token': access_token},
                             data=fp)
            r.raise_for_status()
            callback(model_size, file.stat().st_size)

    # fill zenodo metadata
    data = {'metadata': {
                        'title': mopo_metadata['summary'],
                        'upload_type': 'publication',
                        'publication_type': 'other',
                        'description': markdown(content),
                        'creators': mopo_metadata['authors'],
                        'access_right': 'open',
                        'license': mopo_metadata['license']
                        }
            }

    # add keywords
    if 'tags' in mopo_metadata and mopo_metadata['tags']:
        data['metadata']['keywords'] = mopo_metadata['tags']

    if not private:
        data['metadata']['communities'] = [{'identifier': 'ocr_models'}]

    # add link to training data to metadata
    if 'datasets' in mopo_metadata and mopo_metadata['datasets']:
        data['metadata']['related_identifiers'] = [{'relation': 'isDerivedFrom',
                                                    'identifier': ds,
                                                    'resource_type': 'dataset'} for ds in mopo_metadata['datasets']]

    if 'base_model' in mopo_metadata and mopo_metadata['base_model']:
        if 'related_identifiers' not in data['metadata']:
            data['metadata']['related_identifiers'] = []
        data['metadata']['related_identifiers'].extend([{'relation': 'isDerivedFrom',
                                                         'identifier': mid,
                                                         'resource_type': 'other'} for mid in mopo_metadata['base_model']])

    r = requests.put(f'{get_repo_url()}deposit/depositions/{new_recid}',
                     params={'access_token': access_token},
                     data=json.dumps(data),
                     headers={"Content-Type": "application/json"})

    r.raise_for_status()
    callback(model_size, 1)
    r = requests.post(f'{get_repo_url()}deposit/depositions/{new_recid}/actions/publish',
                      params={'access_token': access_token})
    r.raise_for_status()
    callback(model_size, 1)
    return r.json()['doi']


def publish_model(model: 'PathLike',
                  model_card: str,
                  access_token: str,
                  callback: Callable[[int, int], Any] = lambda total, completed: None,
                  private: bool = False) -> str:
    """
    Publishes a model to the repository.

    Args:
        model: Path to read model from. Can be a single file or a directory.
        model_card: String of a Markdown document with YAML front matter
                    containing the model card + associated metadata.
        access_token: Zenodo API access token
        callback: Function called with octet-wise progress.
        private: Whether to generate a community inclusion request that makes
                 the model recoverable by the public.
    """
    model = Path(model)
    if not model.exists():
        raise ValueError('Model path {model} does not exist.')
    # find all files for the deposition
    model_size = 3
    if model.is_dir():
        model_files = []
        for file in model.iterdir():
            if file.relative_to(model) == Path('README.md'):
                logger.warning('Found a README.md in root of model dir. Ignoring.')
                continue
            if file.is_dir():
                logger.warning(f'Zenodo depositions are flat. Found dir {file}. Skipping')
                continue
            model_files.append(file)
            model_size += file.stat().st_size
    else:
        model_files = [model]
        model_size += model.stat().st_size

    logger.info(f'Got {len(model_files)} files for model with size of {model_size} bytes.')
    # validate model card
    header, content = _yaml_regex.match(model_card).groups()
    mopo_metadata = yaml.safe_load(header)

    # create new deposition
    headers = {"Content-Type": "application/json"}
    r = requests.post(f'{get_repo_url()}deposit/depositions',
                      params={'access_token': access_token},
                      json={},
                      headers=headers)
    r.raise_for_status()
    depo = r.json()
    deposition_id = depo['id']
    bucket_url = depo['links']['bucket']

    # finalize metadata record
    mopo_metadata['id'] = depo['metadata']['prereserve_doi']['doi']
    # we need to validate here to as the data structure might have an empty ID before
    validate(mopo_metadata, _v1_schema, format_checker=format_checker)

    header = yaml.dump(mopo_metadata)

    _metadata = f'---\n{header}---\n{content}'.encode('utf-8')
    model_size += len(_metadata)
    callback(model_size, 1)

    # push metadata file
    r = requests.put(f'{bucket_url}/README.md',
                     params={'access_token': access_token},
                     data=_metadata)
    r.raise_for_status()
    callback(model_size, len(_metadata))
    # push all other files
    for file in model_files:
        with open(file, 'rb') as fp:
            r = requests.put(f'{bucket_url}/{file.name}',
                             params={'access_token': access_token},
                             data=fp)
            r.raise_for_status()
            callback(model_size, file.stat().st_size)

    # fill zenodo metadata
    data = {'metadata': {
                        'title': mopo_metadata['summary'],
                        'upload_type': 'publication',
                        'publication_type': 'other',
                        'description': markdown(content),
                        'creators': mopo_metadata['authors'],
                        'access_right': 'open',
                        'license': mopo_metadata['license']
                        }
            }

    # add keywords
    if 'tags' in mopo_metadata and mopo_metadata['tags']:
        data['metadata']['keywords'] = mopo_metadata['tags']

    if not private:
        data['metadata']['communities'] = [{'identifier': 'ocr_models'}]

    # add link to training data to metadata
    if 'datasets' in mopo_metadata and mopo_metadata['datasets']:
        data['metadata']['related_identifiers'] = [{'relation': 'isDerivedFrom',
                                                    'identifier': ds,
                                                    'resource_type': 'dataset'} for ds in mopo_metadata['datasets']]

    if 'base_model' in mopo_metadata and mopo_metadata['base_model']:
        if 'related_identifiers' not in data['metadata']:
            data['metadata']['related_identifiers'] = []
        data['metadata']['related_identifiers'].extend([{'relation': 'isDerivedFrom',
                                                         'identifier': mid,
                                                         'resource_type': 'other'} for mid in mopo_metadata['base_model']])

    r = requests.put(f'{get_repo_url()}deposit/depositions/{deposition_id}',
                     params={'access_token': access_token},
                     data=json.dumps(data),
                     headers=headers)
    r.raise_for_status()
    callback(model_size, 1)
    r = requests.post(f'{get_repo_url()}deposit/depositions/{deposition_id}/actions/publish',
                      params={'access_token': access_token})
    r.raise_for_status()
    callback(model_size, 1)
    return r.json()['doi']



# -*- coding: utf-8 -*-
import os
import shutil
import tempfile
import unittest
from pathlib import Path
from unittest.mock import patch
from jsonschema.exceptions import ValidationError

thisfile = Path(__file__).resolve().parent
resources = thisfile / 'resources'


@patch.dict(os.environ, {'MODEL_REPO_URL': 'https://sandbox.zenodo.org/api/'})
class TestPublishing(unittest.TestCase):
    """
    Testing interaction with the model repository.
    """

    def setUp(self):
        if (access_token := os.getenv('ZENODO_ACCESS_TOKEN', None)) is None:
            raise unittest.SkipTest('No Zenodo API access token available')
        else:
            self.access_token = access_token

        with open(resources / 'llama_party.md', 'r') as fp:
            self.model_card = fp.read()

        with open(resources / 'invalid_party.md', 'r') as fp:
            self.invalid_model_card = fp.read()

        self.model_path = resources / 'test_model'

    def test_publish(self):
        """
        Tests publishing a model
        """
        from htrmopo import publish_model
        doi = publish_model(model=self.model_path,
                            model_card=self.model_card,
                            access_token=self.access_token)

    def test_invalid_publish(self):
        """
        Tests publishing a model with an invalid model card.
        """
        from htrmopo import publish_model
        with self.assertRaises(ValidationError):
            _ = publish_model(model=self.model_path,
                              model_card=self.invalid_model_card,
                              access_token=self.access_token)

    def test_update(self):
        """
        Tests updating a model
        """
        from htrmopo import update_model, publish_model
        doi = publish_model(model=self.model_path,
                            model_card=self.model_card,
                            access_token=self.access_token)

        doi = update_model(model_id=doi,
                           model=self.model_path,
                           model_card=self.model_card,
                           access_token=self.access_token)

    def test_invalid_metadata_update(self):
        """
        Tests updating a model with an invalid model card.
        """
        from htrmopo import publish_model, update_model
        with self.assertRaises(ValidationError):
            doi = publish_model(model=self.model_path,
                                model_card=self.model_card,
                                access_token=self.access_token)

            _ = update_model(model_id=doi,
                             model=self.model_path,
                             model_card=self.invalid_model_card,
                             access_token=self.access_token)

    def test_invalid_base_record_update(self):
        """
        Tests updating a model with an concept/prior record DOI
        """
        from htrmopo import publish_model, update_model
        with self.assertRaises(ValidationError):
            doi = publish_model(model=self.model_path,
                                model_card=self.model_card,
                                access_token=self.access_token)

            _ = update_model(model_id=doi,
                             model=self.model_path,
                             model_card=self.invalid_model_card,
                             access_token=self.access_token)





import os
import re
import json
import unicodedata

from importlib import resources
from jsonschema import FormatChecker


###############################################################################
###  regex for YAML frontmatter extraction in Markdown files 
###############################################################################

_yaml_delim = r'(?:---|\+\+\+)'
_yaml = r'(.*?)'
_content = r'\s*(.+)$'
_re_pattern = r'^\s*' + _yaml_delim + _yaml + _yaml_delim + _content
_yaml_regex = re.compile(_re_pattern, re.S | re.M)

###############################################################################
###  Schemas
###############################################################################


with open(resources.files('htrmopo').joinpath('v1.metadata.schema.json')) as fp:
    _v1_schema = json.load(fp)

with open(resources.files('htrmopo').joinpath('v0.metadata.schema.json')) as fp:
    _v0_schema = json.load(fp)

###############################################################################
### Controlled vocabularies 
###############################################################################

with open(resources.files('htrmopo').joinpath('iso15924.txt')) as fp:
    _iso15924 = {}
    for line in fp.readlines():
        try:
            code, _, name, *_ = line.split(';')
            _iso15924[code] = name
        except Exception:
            continue

with open(resources.files('htrmopo').joinpath('iso639-3.txt')) as fp:
    _iso639_3 = {}
    for line in fp.readlines()[1:]:
        try:
            code, name, *_ = line.split('\t')
            _iso639_3[code] = name
        except Exception:
            continue

with open(resources.files('htrmopo').joinpath('licenses.json')) as fp:
    _licenses = json.load(fp)


format_checker = FormatChecker()

@format_checker.checks('okfn-license')
def checks_license(instance: object) -> bool:
    if not isinstance(instance, str):
        return True
    return instance in _licenses


@format_checker.checks('iso-639-3')
def checks_iso_639_3(instance: object) -> bool:
    if not isinstance(instance, str):
        return True
    return instance in _iso639_3


@format_checker.checks('iso-15924')
def checks_iso_15924(instance: object) -> bool:
    if not isinstance(instance, str):
        return True
    return instance in _iso15924


###############################################################################
### Global getters 
###############################################################################


def get_oai_url() -> str:
    return os.getenv('MODEL_REPO_OAI_URL', 'https://zenodo.org/oai2d')


def get_repo_url() -> str:
    return os.getenv('MODEL_REPO_URL', 'https://zenodo.org/api/')


###############################################################################
### Mapping between vocabulary IDs and natural language strings 
###############################################################################

def iso639_3_to_name(script: str) -> str:
    return _iso639_3[script]


def iso15924_to_name(script: str) -> str:
    return _iso15924[script]


def license_id_to_name(script: str) -> str:
    return _licenses[script]['title']

###############################################################################
### Mapping between various IDs
###############################################################################

def _doi_to_oai_id(doi: str) -> str:
    """
    Constructs a zenodo OAI identifier from a DOI.

    Args:
        doi: Zenodo DOI
    """
    identifier = re.match(r'[0-9.]+\/zenodo\.([0-9]+)', doi)
    if identifier is not None:
        identifier = identifier.groups()[0]
        return f'oai:zenodo.org:{identifier}'


def _doi_to_zenodo_id(doi: str) -> str:
    """
    Extracts a zenodo record ID from a DOI.

    Args:
        doi: Zenodo DOI
    """
    identifier = re.match(r'[0-9.]+\/zenodo\.([0-9]+)', doi)
    if identifier is not None:
        identifier = identifier.groups()[0]
        return identifier


###############################################################################
### String helpers 
###############################################################################


def is_printable(char: str) -> bool:
    """
    Determines if a chode point is printable/visible when printed.

    Args:
        char: Input code point.

    Returns:
        True if printable, False otherwise.
    """
    letters = ('LC', 'Ll', 'Lm', 'Lo', 'Lt', 'Lu')
    numbers = ('Nd', 'Nl', 'No')
    punctuation = ('Pc', 'Pd', 'Pe', 'Pf', 'Pi', 'Po', 'Ps')
    symbol = ('Sc', 'Sk', 'Sm', 'So')
    printable = letters + numbers + punctuation + symbol

    return unicodedata.category(char) in printable


def make_printable(char: str) -> str:
    """
    Takes a Unicode code point and return a printable representation of it.

    Args:
        char: Input code point

    Returns:
        Either the original code point, the name of the code point if it is a
        combining mark, whitespace etc., or the hex code if it is a control
        symbol.
    """
    if not char or is_printable(char):
        return char
    elif unicodedata.category(char) in ('Cc', 'Cs', 'Co'):
        return '0x{:x}'.format(ord(char))
    else:
        return unicodedata.name(char)



#
# Copyright 2024 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
DCAT record class for Sickle
"""
from sickle.models import Header, Record

from dataclasses import dataclass, field
from typing import TypedDict, Optional, Dict, List, TYPE_CHECKING

if TYPE_CHECKING:
    import datetime.datetime


def dcat_to_dict(tree):
    """Converts a DCAT XML tree to a dictionary.
    """
    title = t.text if (t := tree.find('./{*}RDF/{*}Description/{*}title')) is not None else None

    distribution = []
    for element in tree.findall('./{*}RDF/{*}Description/{*}distribution'):
        if (file_url := element.find('./{*}Distribution/{*}downloadURL')) is not None:
            distribution.append({'url': file_url.attrib.values()[0],
                                 'size': int(siz.text) if (siz := element.find('./{*}Distribution/{*}byteSize')) is not None else -1})
    if (description := tree.find('./{*}RDF/{*}Description/{*}description')) is not None:
        description = description.text
    keywords = [keyword.text for keyword in tree.findall('./{*}RDF/{*}Description/{*}keyword')]
    creators = []
    for creator in tree.findall('./{*}RDF/{*}Description/{*}creator/{*}Description'):
        creators.append({'name': c.text if (c := creator.find('./{*}name')) is not None else '',
                         'orcid': o[0] if len(o := [v for k, v in creator.attrib.items() if k.endswith('about')]) else None,
                         'affiliation': a.text if (a := creator.find('./{*}memberOf//{*}name')) is not None else ''})
    if (concept := tree.find('./{*}RDF/{*}Description/{*}isVersionOf/{*}Description/{*}identifier')) is not None:
        concept = concept.text
    return {'doi': d.text if (d := tree.find('./{*}RDF/{*}Description/{*}identifier')) is not None else None,
            'summary': title,
            'description': description,
            'distribution': distribution,
            'creators': creators,
            'keywords': keywords,
            'concept_doi': concept
            }


class DCATRecord(Record):
    """Represents an OAI DCAT record.

    :param record_element: The XML element 'record'.
    :type record_element: :class:`lxml.etree._Element`
    :param strip_ns: Flag for whether to remove the namespaces from the
                     element names.
    """

    def __init__(self, record_element, strip_ns=True):
        super(Record, self).__init__(record_element, strip_ns=strip_ns)
        self.header = Header(self.xml.find(
            './/' + self._oai_namespace + 'header'))
        self.deleted = self.header.deleted
        if not self.deleted:
            self.metadata = self.get_metadata()

    def __repr__(self):
        if self.header.deleted:
            return '<Record %s [deleted]>' % self.header.identifier
        else:
            return '<Record %s>' % self.header.identifier

    def __iter__(self):
        return iter(self.metadata.items())

    def get_metadata(self):
        return dcat_to_dict(self.xml.find('.//{*}metadata'))


class AuthorDict(TypedDict):
    author: str
    affiliation: Optional[str]
    orcid: Optional[str]


class DistributionDict(TypedDict):
    url: str
    size: int


class v0MetricsDict(TypedDict):
    cer: float


@dataclass
class v1RepositoryRecord:
    """
    A v1 record in the repository representing a single ATR model.
    """
    doi: str
    concept_doi: str
    creators: List[AuthorDict]
    summary: str
    description: str
    license: str
    software_name: str
    script: List[str]
    language: List[str]
    publication_date: 'datetime.datetime'
    distribution: List[DistributionDict]
    model_type: List[str]
    license_url: Optional[str] = None
    software_hints: Optional[List[str]] = None
    model_type: Optional[List[str]] = None
    metrics: Optional[Dict[str, float]] = None
    datasets: Optional[List[str]] = None
    base_model: Optional[List[str]] = None
    keywords: Optional[List[str]] = None
    publication: Optional[str] = None
    version: str = 'v1'


@dataclass
class v0RepositoryRecord:
    """
    A v0 record in the repository representing a single ATR model.
    """
    doi: str
    concept_doi: str
    creators: List[AuthorDict]
    summary: str
    description: str
    metrics: v0MetricsDict
    license: str
    script: List[str]
    distribution: List[DistributionDict]
    graphemes: List[str]
    publication_date: 'datetime.datetime'
    model_type: List[str] = field(default_factory=lambda: ['recognition'])
    keywords: Optional[List[str]] = None
    version: str = 'v0'



#
# Copyright 2025 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
htrmopo.cli
~~~~~~~~~~~

Command line driver for interacting with the model repository.
"""
import logging

import click

from rich import print
from rich.tree import Tree
from rich.table import Table
from rich.console import Group
from rich.traceback import install
from rich.logging import RichHandler
from rich.markdown import Markdown
from rich.progress import Progress

from collections import defaultdict

from htrmopo.util import make_printable, is_printable, iso15924_to_name, iso639_3_to_name, license_id_to_name

logging.captureWarnings(True)
logger = logging.getLogger('htrmopo')


# install rich traceback handler
install(suppress=[click])


def message(msg: str, **styles) -> None:
    if logger.getEffectiveLevel() >= 30:
        click.secho(msg, **styles)


def _render_creators(creators):
    o = []
    for creator in creators:
        c_text = creator['name']
        if (orcid := creator.get('orcid', None)) is not None:
            c_text += f' ({orcid})'
        if (affiliation := creator.get('affiliation', None)) is not None:
            c_text += f' ({affiliation})'
        o.append(c_text)
    return o


def _render_metrics(metrics):
    return [f'{k}: {v:.2f}' for k, v in metrics.items()]


@click.group(chain=False)
@click.version_option()
@click.option('-v', '--verbose', default=0, count=True, show_default=True)
def cli(verbose):
    """
    Base command for repository interaction.
    """
    ctx = click.get_current_context()
    ctx.meta['verbose'] = verbose

    logger.addHandler(RichHandler(rich_tracebacks=True))
    logger.setLevel(level=30 - min(10 * verbose, 20))


@cli.command('show')
@click.pass_context
@click.option('-V', '--metadata-version',
              default='highest',
              type=click.Choice(['v0', 'v1', 'highest']),
              help='Version of metadata to fetch if multiple exist in repository.')
@click.argument('model_id')
def show(ctx, metadata_version, model_id):
    """
    Retrieves a model description from the repository.
    """
    from htrmopo import get_description

    if metadata_version == 'highest':
        metadata_version = None

    try:
        desc = get_description(model_id, version=metadata_version)
    except ValueError as e:
        logger.error(e)
        ctx.exit(1)

    if desc.version == 'v0':
        chars = []
        combining = []
        for char in sorted(desc.graphemes):
            if not is_printable(char):
                combining.append(make_printable(char))
            else:
                chars.append(char)

        table = Table(title=desc.summary, show_header=False)
        table.add_column('key', justify="left", no_wrap=True)
        table.add_column('value', justify="left", no_wrap=False)
        table.add_row('DOI', desc.doi)
        table.add_row('concept DOI', desc.concept_doi)
        table.add_row('publication date', desc.publication_date.isoformat())
        table.add_row('model type', Group(*desc.model_type))
        table.add_row('script', Group(*[iso15924_to_name(x) for x in desc.script]))
        table.add_row('alphabet', Group(' '.join(chars), ', '.join(combining)))
        table.add_row('keywords', Group(*desc.keywords))
        table.add_row('metrics', Group(*_render_metrics(desc.metrics)))
        table.add_row('license', license_id_to_name(desc.license))
        table.add_row('creators', Group(*_render_creators(desc.creators)))
        table.add_row('description', desc.description)
    elif desc.version == 'v1':
        table = Table(title=desc.summary, show_header=False)
        table.add_column('key', justify="left", no_wrap=True)
        table.add_column('value', justify="left", no_wrap=False)
        table.add_row('DOI', desc.doi)
        table.add_row('concept DOI', desc.concept_doi)
        table.add_row('publication date', desc.publication_date.isoformat())
        table.add_row('model type', Group(*desc.model_type))
        table.add_row('language', Group(*[iso639_3_to_name(x) for x in desc.language]))
        table.add_row('script', Group(*[iso15924_to_name(x) for x in desc.script]))
        table.add_row('keywords', Group(*desc.keywords))
        table.add_row('datasets', Group(*desc.datasets))
        table.add_row('metrics', Group(*_render_metrics(desc.metrics)))
        table.add_row('base model', Group(*desc.base_model))
        table.add_row('software', desc.software_name)
        table.add_row('software_hints', Group(*desc.software_hints))
        table.add_row('license', license_id_to_name(desc.license))
        table.add_row('creators', Group(*_render_creators(desc.creators)))
        table.add_row('description', Markdown(desc.description))

    print(table)


@cli.command('list')
@click.pass_context
@click.option('-F', '--from-date',
              default=None,
              type=click.STRING,
              help='ISO-8601 date string to filter repository entries by age.')
def list_models(ctx, from_date):
    """
    Lists models in the repository.
    """
    from htrmopo import get_listing

    with Progress() as progress:
        download_task = progress.add_task('Retrieving model list', total=0, visible=True)
        repository = get_listing(lambda total, advance: progress.update(download_task, total=total, advance=advance), from_date=from_date)
    # aggregate models under their concept DOI
    concepts = defaultdict(list)
    for item in repository.values():
        # both got the same DOI information
        record = item['v0'] if item['v0'] else item['v1']
        concepts[record.concept_doi].append(record.doi)

    table = Table(show_header=True)
    table.add_column('DOI', justify="left", no_wrap=True)
    table.add_column('summary', justify="left", no_wrap=False)
    table.add_column('model type', justify="left", no_wrap=False)
    table.add_column('keywords', justify="left", no_wrap=False)

    for k, v in concepts.items():
        records = [repository[x]['v1'] if 'v1' in repository[x] else repository[x]['v0'] for x in v]
        records = sorted(records, key=lambda x: x.publication_date, reverse=True)
        t = Tree(k)
        [t.add(x.doi) for x in records]
        table.add_row(t,
                      Group(*[''] + [x.summary for x in records]),
                      Group(*[''] + ['; '.join(x.model_type) for x in records]),
                      Group(*[''] + ['; '.join(x.keywords) for x in records]))

    print(table)
    ctx.exit(0)


@cli.command('get')
@click.pass_context
@click.option('-o', '--output', show_default=True,
              type=click.Path(file_okay=False, writable=True), help='Model card file for the model.')
@click.argument('model_id')
def get(ctx, output, model_id):
    """
    Retrieves a model from the repository.
    """
    from htrmopo import get_model

    with Progress() as progress:
        download_task = progress.add_task('Processing', total=0, visible=True)
        filename = get_model(model_id,
                             output,
                             lambda total, advance: progress.update(download_task, total=total, advance=advance))
    message(f'Model name: {filename}')
    ctx.exit(0)


@cli.command('publish')
@click.pass_context
@click.option('-i', '--metadata', show_default=True,
              type=click.File(mode='r', lazy=True), help='Model card file for the model.')
@click.option('-d', '--doi', help='DOI of an existing model. If set the record will be updated by creating a new version.')
@click.option('-a', '--access-token', prompt=True, help='Zenodo access token')
@click.option('-p', '--private/--public', default=False, help='Disables Zenodo '
              'community inclusion request. Allows upload of models that will not show '
              'up on `htrmopo list` output')
@click.argument('model', nargs=1, type=click.Path(exists=False, readable=True, dir_okay=True))
def publish(ctx, metadata, doi, access_token, private, model):
    """
    Publishes a model on the Zenodo model repository.
    """
    from htrmopo import publish_model, update_model

    pub_fn = publish_model

    with Progress() as progress:
        upload_task = progress.add_task('Uploading', total=0)
        kwargs = {'model': model,
                  'model_card': metadata.read(),
                  'access_token': access_token,
                  'callback': lambda total, advance: progress.update(upload_task, total=total, advance=advance),
                  'private': private}
        if doi:
            pub_fn = update_model
            kwargs['model_id'] = doi
        oid = pub_fn(**kwargs)
    print(f'model PID: {oid}')
